class AES{
	constructor(){
		this.password = Utilities.hexToBase62(Utilities.generatePassword(16)); //16 bytes
		this.salt = Utilities.hexToBase62(Utilities.generatePassword(16)); //16 bytes
		this.rounds = 10000; //Of PBKDF2, this is very low for dev speed
		console.log('generated password for AES encryption',this.password);
		this.E = new Encrypter(this.password,this.salt,this.rounds);
		this.D = new Decrypter(this.password,this.rounds);
	}
	setPassword(password){
		this.password = password;
		this.E = new Encrypter(password,this.salt,this.rounds);
		this.D = new Decrypter(password,this.rounds);
	}
	async encrypt(text){ //Returns ciphertoken
		return this.E.encryptAsCiphertoken(text);
	}
	async encryptFile(text){ //Returns ciphertoken
		return this.E.encryptFileAsCiphertoken(text);
	}
	async decrypt(ciphertoken){ //Returns text
		return this.D.decrypt(ciphertoken);
	}
	async decryptFile(ciphertoken){ //Returns text
		return this.D.decryptFile(ciphertoken);
	}
}

class Encrypter{
	constructor(password,salt,roundsOfPBKDF2){
		if(typeof roundsOfPBKDF2 =='undefined' || roundsOfPBKDF2 == null)
			this.roundsOfPBKDF2=1000;
		else
			this.roundsOfPBKDF2=roundsOfPBKDF2; //Consider mobile...
		this.salt = salt;
		this.CryptoKey = this.generateCryptoKey(password);
	}
	async generateCryptoKey(password){
		const passwordBytes = new TextEncoder('utf-8').encode(password);
		const salt = await window.crypto.getRandomValues(new Uint8Array(16));
		this.salt = salt;
		const CryptoKey = await window.crypto.subtle.importKey(
			"raw",
			passwordBytes,
			{"name": "PBKDF2"},
			false,
			["deriveKey"]
		);
		
		return window.crypto.subtle.deriveKey(
			{
				"name": "PBKDF2",
				"salt": salt,
				"iterations": this.roundsOfPBKDF2,
				"hash": 'SHA-256'
			},
			CryptoKey,
			{"name": "AES-GCM", "length": 256},
			false,
			["encrypt", "decrypt"]
		);
	}
	async encryptFileAsCiphertoken(buffer){
		const CryptoKey = await this.CryptoKey;
		const iv = await window.crypto.getRandomValues(new Uint8Array(16));
		const encryptedBuffer = await window.crypto.subtle.encrypt(
			{name: "AES-GCM", iv: iv},
			CryptoKey,
			buffer
		);
		
		return [
			Utilities.toHex(this.salt),
			Utilities.toHex(iv),
			Utilities.toHex(new Uint8Array(encryptedBuffer))
		].join(','); //CSV
	}
	
	async encryptAsCiphertoken(msg){
		const CryptoKey = await this.CryptoKey;
		const encoded = new TextEncoder("utf-8").encode(msg);
		const iv = await window.crypto.getRandomValues(new Uint8Array(16));
		const encryptedBuffer = await window.crypto.subtle.encrypt(
			{name: "AES-GCM", iv: iv},
			CryptoKey,
			encoded
		);
		
		
		return [
			Utilities.toHex(this.salt),
			Utilities.toHex(iv),
			Utilities.toHex(new Uint8Array(encryptedBuffer))
		].join(','); //CSV
	}
	
}
class Decrypter{
	constructor(password,roundsOfPBKDF2){
		this.password = password;
		if(typeof roundsOfPBKDF2 =='undefined' || roundsOfPBKDF2 == null)
			this.roundsOfPBKDF2=1000;
		else
			this.roundsOfPBKDF2=roundsOfPBKDF2; //Consider mobile...
	}
	async decryptFile(ciphertoken){
		const parsedCiphertoken = this.parseCiphertoken(ciphertoken);
		
		const CryptoKey = await this.generateCryptoKey(this.password,parsedCiphertoken.salt);
		
		const buffer = await window.crypto.subtle.decrypt(
			{
				name: "AES-GCM",
				iv: parsedCiphertoken.iv,
			},
			CryptoKey,
			parsedCiphertoken.encryptedBytes
		);
		return buffer;
	}
	async decrypt(ciphertoken){
		const parsedCiphertoken = this.parseCiphertoken(ciphertoken);
		
		const CryptoKey = await this.generateCryptoKey(this.password,parsedCiphertoken.salt);
		
		const buffer = await window.crypto.subtle.decrypt(
			{
				name: "AES-GCM",
				iv: parsedCiphertoken.iv,
			},
			CryptoKey,
			parsedCiphertoken.encryptedBytes
		);
		//console.log('finished decryption step resulting in this buffer:',buffer);
		if(buffer != false){
			const decoded = new TextDecoder('utf-8').decode(buffer);
			return decoded;
		}
		return false;
	}
	parseCiphertoken(ciphertoken){
		if(ciphertoken.length < 3)
			throw new Error('Invalid ciphertoken. Too short.',ciphertoken);
		if(ciphertoken.indexOf(',')==-1)
			throw new Error('ciphertoken must contain a comma');
		
		const components = ciphertoken.split(',');
		return {
			'salt':Utilities.fromHex(components[0]),
			'iv':Utilities.fromHex(components[1]),
			'encryptedBytes':Utilities.fromHex(components[2]),
		};
	}
	async generateCryptoKey(password,salt){
		const passwordBytes = new TextEncoder('utf-8').encode(password);
		const CryptoKey = await window.crypto.subtle.importKey(
			"raw",
			passwordBytes,
			{"name": "PBKDF2"},
			false,
			["deriveKey"]
		);
		
		
		return window.crypto.subtle.deriveKey(
			{
				"name": "PBKDF2",
				"salt": salt,
				"iterations": this.roundsOfPBKDF2,
				"hash": 'SHA-256'
			},
			CryptoKey,
			{"name": "AES-GCM", "length": 256},
			false,
			["encrypt", "decrypt"]
		);
	}
}
