class Utilities{
	static getPBKDF2Params(salt){ //salt=uint8
		const iterations = 1000000;
		return { name: 'PBKDF2', hash: 'SHA-256', salt: salt, iterations: iterations};
	}
	static toHex(bytes){
		if (!bytes) //Shouldn't occur
			return null;
		else if(typeof bytes != 'object')
			throw new Error('Parameter must be of type Uint8Array (object)');

		var hexChars = [];
		for(var i=0;i<bytes.length;i++){
			var hexChar = bytes[i].toString(16);
			hexChars.push(hexChar.length < 2 ? "0" + hexChar:hexChar);
		}
		return hexChars.join("");
	}
	static fromHex(hexString){
		if(typeof hexString != 'string')
			throw new Error("fromHex can only be used on data of type string.");
		else if(hexString.length % 2 != 0)
			throw new Error("Invalid hex-encoded string, the length must be even.");

		var arrayBuffer = new Uint8Array(hexString.length / 2);
		for(var i=0;i<hexString.length;i+=2)
			arrayBuffer[i/2] = parseInt(hexString.substr(i, 2), 16);
		return arrayBuffer;
	};
	

	static async hash(stringIn){
		return await crypto.subtle.digest('SHA-256',new TextEncoder('utf-8').encode(stringIn)).then(bytes=>{
			return Utilities.toHex(new Uint8Array(bytes));
		});
	}
	static async generatePBKDF2(stringIn,salt){
		const stringInEncoded = new TextEncoder().encode(stringIn);
		const key = await crypto.subtle.importKey('raw', stringInEncoded, 'PBKDF2', false, ['deriveBits']);

		if(typeof salt == 'undefined' || salt.length==0)
			salt = crypto.getRandomValues(new Uint8Array(16));

		const params = Utilities.getPBKDF2Params(salt);
		const bytes = await crypto.subtle.deriveBits(params, key, 256);
		return Utilities.toHex(new Uint8Array(bytes))+"."+Utilities.toHex(salt); //Generate bytes.salt
	}
	static async verifyPBKDF2(PBKDF2,stringIn){ //PBKDF2 = bytes.salt
		var temp = PBKDF2.split('.');
		const expectedHash = temp[0];
		var salt = temp[1];
		if(salt.length < 8)
			throw new Error('Invalid PBKDF2');
		else
			salt = Utilities.fromHex(salt);
		temp = await Utilities.generatePBKDF2(stringIn,salt);
		temp = temp.split('.');
		const actualHash = temp[0];
		if(actualHash === expectedHash)
			return true;
		return false;
	}
	static generatePassword(byteLength){
		if(typeof byteLength != 'number')
			throw new Error('byteLength must be a number');
		else if(byteLength < 12)
			throw new Error('byteLength must be at least 12 bytes (not '+byteLength+'). Anything less is uselessly insecure.');
		const bytes = window.crypto.getRandomValues(new Uint8Array(byteLength));
		return Utilities.toHex(bytes);
	}
	static hexToBase62(hex){
		//hex = "0x"+hex;
		//const b = BigInt(hex);
		const h = bigInt(hex,16);
		const b62 = Utilities.toBase62(h);
		//console.log('b62',b62);
		return b62;
	}
	static toBase62(nIn){
		var base62 = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'.split('');
		var output = '';
		//console.log('nIn',nIn);
		const b = bigInt(62);
		for(let i=100;i>0;i--){ //Max of 64 digits long
			//Below is what it should be but BigInt doesn't work on mobile safari and Chrome...
			const r = nIn.divmod(b);
			nIn = r.quotient;
			const remainder = r.remainder;
			/*
			const remainder = nIn % b;
			nIn = nIn/b;
			*/
			if(remainder == 0 && nIn == 0)
				break;

			output = base62[remainder]+""+output;
		}
		return output;
	}
	
}
